"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    NoReturn,
)

from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    PostgreSqlMinosDatabase,
)


class HandlerSetup(PostgreSqlMinosDatabase):
    """Minos Broker Setup Class"""

    TABLE_NAME: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def _setup(self) -> NoReturn:
        await self._create_event_queue_table()

    async def _create_event_queue_table(self) -> NoReturn:
        _CREATE_TABLE_QUERY = SQL(
            "CREATE TABLE IF NOT EXISTS {} ("
            '"id" BIGSERIAL NOT NULL PRIMARY KEY, '
            '"topic" VARCHAR(255) NOT NULL, '
            '"partition_id" INTEGER,'
            '"binary_data" BYTEA NOT NULL, '
            '"retry" INTEGER NOT NULL DEFAULT 0,'
            '"creation_date" TIMESTAMP NOT NULL)'
        )
        await self.submit_query(_CREATE_TABLE_QUERY.format(Identifier(self.TABLE_NAME)), lock=hash(self.TABLE_NAME))



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
__version__ = "0.0.9"

from .brokers import (
    Broker,
    BrokerSetup,
    CommandBroker,
    CommandReplyBroker,
    EventBroker,
    Producer,
    ProducerService,
)
from .discovery import (
    DiscoveryConnector,
    MinosDiscoveryClient,
)
from .exceptions import (
    MinosActionNotFoundException,
    MinosDiscoveryConnectorException,
    MinosNetworkException,
)
from .handlers import (
    CommandConsumer,
    CommandConsumerService,
    CommandHandler,
    CommandHandlerService,
    CommandReplyConsumer,
    CommandReplyConsumerService,
    CommandReplyHandler,
    CommandReplyHandlerService,
    CommandRequest,
    CommandResponse,
    CommandResponseException,
    Consumer,
    EventConsumer,
    EventConsumerService,
    EventHandler,
    EventHandlerService,
    Handler,
    HandlerEntry,
    HandlerSetup,
)
from .rest import (
    HttpRequest,
    HttpResponse,
    HttpResponseException,
    RestBuilder,
    RestService,
)
from .snapshots import (
    SnapshotService,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .builders import (
    RestBuilder,
)
from .messages import (
    HttpRequest,
    HttpResponse,
    HttpResponseException,
)
from .services import (
    RestService,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from __future__ import (
    annotations,
)

import logging
import socket
from typing import (
    NoReturn,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
)

from .clients import (
    MinosDiscoveryClient,
)

logger = logging.getLogger(__name__)


class DiscoveryConnector(MinosSetup):
    """Discovery Connector class."""

    def __init__(self, client, name: str, host: str, port: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client = client

        self.name = name
        self.host = host
        self.port = port

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> DiscoveryConnector:
        client = MinosDiscoveryClient(host=config.discovery.host, port=config.discovery.port)
        port = config.rest.broker.port
        name = config.service.name
        host = socket.gethostbyname(socket.getfqdn())
        return cls(client=client, name=name, host=host, port=port, *args, **kwargs)

    async def _setup(self) -> NoReturn:
        await self.subscribe()

    async def subscribe(self) -> NoReturn:
        """Send a subscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery subscription...")
        await self.client.subscribe(self.host, self.port, self.name)

    async def _destroy(self) -> NoReturn:
        await self.unsubscribe()

    async def unsubscribe(self) -> NoReturn:
        """Send an unsubscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery unsubscription...")
        await self.client.unsubscribe(self.name)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
)

from minos.common import (
    Command,
    MinosConfig,
)

from .abc import (
    Broker,
)

logger = logging.getLogger(__name__)


class CommandBroker(Broker):
    """Minos Command Broker Class."""

    ACTION = "command"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandBroker:
        return cls(*args, **config.commands.queue._asdict(), **kwargs)

    # noinspection PyMethodOverriding
    async def send(self, data: Any, topic: str, saga: str, reply_topic: str, **kwargs) -> int:
        """Send a ``Command``.

        :param data: The data to be send.
        :param topic: Topic in which the message will be published.
        :param saga: Saga identifier.
        :param reply_topic: Topic name in which the reply will be published.
        :return: This method does not return anything.
        """
        command = Command(topic, data, saga, reply_topic)
        logger.info(f"Sending '{command!s}'...")
        return await self.send_bytes(command.topic, command.avro_bytes)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .clients import (
    MinosDiscoveryClient,
)
from .connectors import (
    DiscoveryConnector,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
)
from uuid import (
    UUID,
)

from minos.common import (
    CommandReply,
    CommandStatus,
    MinosConfig,
)

from .abc import (
    Broker,
)

logger = logging.getLogger(__name__)


class CommandReplyBroker(Broker):
    """Minos Command Broker Class."""

    ACTION = "commandReply"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandReplyBroker:
        return cls(*args, **config.saga.queue._asdict(), **kwargs)

    # noinspection PyMethodOverriding
    async def send(self, data: Any, topic: str, saga: UUID, status: CommandStatus, **kwargs) -> int:
        """Send a ``CommandReply``.

        :param data: The data to be send.
        :param topic: Topic in which the message will be published.
        :param saga: Saga identifier.
        :param status: Command status.
        :return: This method does not return anything.
        """

        command_reply = CommandReply(f"{topic}Reply", data, saga, status)
        logger.info(f"Sending '{command_reply!s}'...")
        return await self.send_bytes(command_reply.topic, command_reply.avro_bytes)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .consumers import (
    CommandConsumer,
)
from .handlers import (
    CommandHandler,
)
from .messages import (
    CommandRequest,
    CommandResponse,
    CommandResponseException,
)
from .services import (
    CommandConsumerService,
    CommandHandlerService,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from minos.common import (
    MinosException,
)


class MinosNetworkException(MinosException):
    """Base network exception."""


class MinosDiscoveryConnectorException(MinosNetworkException):
    """Exception to be raised when there is a failure while communicating with the discovery."""


class MinosActionNotFoundException(MinosNetworkException):
    """Exception to be raised when an action cannot be found,"""



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    NoReturn,
)

from aiokafka import (
    AIOKafkaProducer,
)
from psycopg2.sql import (
    SQL,
)

from minos.common import (
    BROKER,
    QUEUE,
    MinosConfig,
)

from .abc import (
    BrokerSetup,
)

logger = logging.getLogger(__name__)


class Producer(BrokerSetup):
    """Minos Queue Dispatcher Class."""

    # noinspection PyUnresolvedReferences
    def __init__(self, *args, queue: QUEUE, broker: BROKER, **kwargs):
        # noinspection PyProtectedMember
        super().__init__(*args, **queue._asdict(), **kwargs)
        self.retry = queue.retry
        self.records = queue.records
        self.broker = broker

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> Producer:
        return cls(*args, **config.events._asdict(), **kwargs)

    async def dispatch(self) -> NoReturn:
        """Dispatch the items in the publishing queue.

        :return: This method does not return anything.
        """
        pool = await self.pool
        with await pool.cursor() as cursor:
            # aiopg works in autocommit mode, meaning that you have to use transaction in manual mode.
            # Read more details: https://aiopg.readthedocs.io/en/stable/core.html#transactions.
            await cursor.execute("BEGIN")

            # Select records and lock them FOR UPDATE
            await cursor.execute(_SELECT_NON_PROCESSED_ROWS_QUERY, (self.retry, self.records))
            result = await cursor.fetchall()

            for row in result:
                published = False
                try:
                    published = await self.publish(topic=row[1], message=row[2])
                except Exception as exc:  # pragma: no cover
                    logger.warning(f"Raised an exception while publishing a message: {exc!r}")
                finally:
                    if published:
                        await cursor.execute(_DELETE_PROCESSED_QUERY, (row[0],))
                    else:
                        await cursor.execute(_UPDATE_NON_PROCESSED_QUERY, (row[0],))

            # Manually commit
            await cursor.execute("COMMIT")

    async def publish(self, topic: str, message: bytes) -> bool:
        """Publish a new item in the broker (kafka).

        :param topic: The topic in which the message will be published.
        :param message: The message to be published.
        :return: A boolean flag, ``True`` when the message is properly published or ``False`` otherwise.
        """
        logger.debug(f"Producing message with {topic!s} topic...")

        producer = AIOKafkaProducer(bootstrap_servers=f"{self.broker.host}:{self.broker.port}")
        # noinspection PyBroadException
        try:
            # Get cluster layout and initial topic/partition leadership information
            await producer.start()
            # Produce message
            await producer.send_and_wait(topic, message)
            published = True
        except Exception:
            published = False
        finally:
            # Wait for all pending messages to be delivered or expire.
            await producer.stop()

        return published


_SELECT_NON_PROCESSED_ROWS_QUERY = SQL(
    "SELECT * "
    "FROM producer_queue "
    "WHERE retry < %s "
    "ORDER BY creation_date "
    "LIMIT %s "
    "FOR UPDATE "
    "SKIP LOCKED"
)

_DELETE_PROCESSED_QUERY = SQL("DELETE FROM producer_queue " "WHERE id = %s")

_UPDATE_NON_PROCESSED_QUERY = SQL("UPDATE producer_queue " "SET retry = retry + 1 " "WHERE id = %s")



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import logging
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Awaitable,
    Callable,
    NoReturn,
    Optional,
    Tuple,
    Union,
)

from dependency_injector.wiring import (
    Provide,
)

from minos.common import (
    Command,
    CommandStatus,
    MinosBroker,
    MinosConfig,
    MinosException,
    Response,
    ResponseException,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)
from .messages import (
    CommandRequest,
)

logger = logging.getLogger(__name__)


class CommandHandler(Handler):
    """Command Handler class."""

    TABLE_NAME = "command_queue"
    ENTRY_MODEL_CLS = Command

    broker: MinosBroker = Provide["command_reply_broker"]

    def __init__(self, broker: MinosBroker = None, **kwargs: Any):
        super().__init__(**kwargs)

        if broker is not None:
            self.broker = broker

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandHandler:
        handlers = {item.name: {"controller": item.controller, "action": item.action} for item in config.commands.items}
        return cls(handlers=handlers, **config.commands.queue._asdict(), **kwargs)

    async def dispatch_one(self, entry: HandlerEntry) -> NoReturn:
        """Dispatch one row.

        :param entry: Entry to be dispatched.
        :return: This method does not return anything.
        """
        command = entry.data
        logger.info(f"Dispatching '{command!s}'...")

        fn = self.get_callback(entry.callback)
        items, status = await fn(command)

        await self.broker.send(items, topic=command.reply_topic, saga=command.saga, status=status)

    @staticmethod
    def get_callback(
        fn: Callable[[CommandRequest], Union[Optional[CommandRequest], Awaitable[Optional[CommandRequest]]]]
    ) -> Callable[[Command], Awaitable[Tuple[Any, CommandStatus]]]:
        """Get the handler function to be used by the Command Handler.

        :param fn: The action function.
        :return: A wrapper function around the given one that is compatible with the Command Handler API.
        """

        async def _fn(command: Command) -> Tuple[Any, CommandStatus]:
            try:
                request = CommandRequest(command)
                response = fn(request)
                if isawaitable(response):
                    response = await response
                if isinstance(response, Response):
                    response = await response.content()
                return response, CommandStatus.SUCCESS
            except ResponseException as exc:
                logger.info(f"Raised a user exception: {exc!s}")
                return None, CommandStatus.ERROR
            except MinosException as exc:
                logger.warning(f"Raised a 'minos' exception: {exc!r}")
                return None, CommandStatus.SYSTEM_ERROR
            except Exception as exc:
                logger.exception(f"Raised an exception: {exc!r}.")
                return None, CommandStatus.SYSTEM_ERROR

        return _fn



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.

from __future__ import (
    annotations,
)

import logging
from abc import (
    abstractmethod,
)
from datetime import (
    datetime,
)
from inspect import (
    isclass,
)
from typing import (
    Any,
    Callable,
    NoReturn,
    Type,
)

from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    MinosModel,
    import_module,
)

from ...exceptions import (
    MinosActionNotFoundException,
)
from ..entries import (
    HandlerEntry,
)
from .setups import (
    HandlerSetup,
)

logger = logging.getLogger(__name__)


class Handler(HandlerSetup):
    """
    Event Handler

    """

    __slots__ = "_handlers", "_records", "_retry"

    ENTRY_MODEL_CLS: Type[MinosModel]

    def __init__(self, records: int, handlers: dict[str, dict[str, Any]], retry: int, **kwargs: Any):
        super().__init__(**kwargs)
        self._handlers = handlers
        self._records = records
        self._retry = retry

    async def dispatch(self) -> NoReturn:
        """Event Queue Checker and dispatcher.

        It is in charge of querying the database and calling the action according to the topic.

            1. Get periodically 10 records (or as many as defined in config > queue > records).
            2. Instantiate the action (asynchronous) by passing it the model.
            3. If the invoked function terminates successfully, remove the event from the database.

        Raises:
            Exception: An error occurred inserting record.
        """

        pool = await self.pool
        with await pool.cursor() as cursor:
            # aiopg works in autocommit mode, meaning that you have to use transaction in manual mode.
            # Read more details: https://aiopg.readthedocs.io/en/stable/core.html#transactions.
            await cursor.execute("BEGIN")

            # Select records and lock them FOR UPDATE
            await cursor.execute(
                _SELECT_NON_PROCESSED_ROWS_QUERY.format(Identifier(self.TABLE_NAME)), (self._retry, self._records)
            )
            result = await cursor.fetchall()

            for row in result:
                dispatched = await self._dispatch_one(row)
                if dispatched:
                    await cursor.execute(_DELETE_PROCESSED_QUERY.format(Identifier(self.TABLE_NAME)), (row[0],))
                else:
                    await cursor.execute(_UPDATE_NON_PROCESSED_QUERY.format(Identifier(self.TABLE_NAME)), (row[0],))

            # Manually commit
            await cursor.execute("COMMIT")

    async def _dispatch_one(self, row: tuple) -> bool:
        try:
            entry = await self._build_entry(row)
        except Exception as exc:
            logger.warning(f"Raised an exception while building the message with id={row[0]}: {exc!r}")
            return False

        logger.debug(f"Dispatching '{entry.data!s}'...")

        try:
            await self.dispatch_one(entry)
        except Exception as exc:
            logger.warning(f"Raised an exception while dispatching {entry.data!s}: {exc!r}")
            return False

        return True

    async def _build_entry(self, row: tuple[int, str, int, bytes, datetime]) -> HandlerEntry:
        id = row[0]
        topic = row[1]
        callback = self.get_action(row[1])
        partition_id = row[2]
        data = self.ENTRY_MODEL_CLS.from_avro_bytes(row[3])
        retry = row[4]
        created_at = row[5]

        entry = HandlerEntry(id, topic, callback, partition_id, data, retry, created_at)
        return entry

    def get_action(self, topic: str) -> Callable:
        """Get Event instance to call.

        Gets the instance of the class and method to call.

        Args:
            topic: Kafka topic. Example: "TicketAdded"

        Raises:
            MinosNetworkException: topic TicketAdded have no controller/action configured, please review th
                configuration file.
        """
        if topic not in self._handlers:
            raise MinosActionNotFoundException(
                f"topic {topic} have no controller/action configured, " f"please review th configuration file"
            )

        event = self._handlers[topic]

        controller = import_module(event["controller"])
        if isclass(controller):
            controller = controller()
        action = getattr(controller, event["action"])

        logger.debug(f"Loaded {action!r} action!")
        return action

    @abstractmethod
    async def dispatch_one(self, entry: HandlerEntry) -> NoReturn:
        """Dispatch one row.

        :param entry: Entry to be dispatched.
        :return: This method does not return anything.
        """
        raise NotImplementedError


_SELECT_NON_PROCESSED_ROWS_QUERY = SQL(
    "SELECT * FROM {} WHERE retry < %s ORDER BY creation_date LIMIT %s FOR UPDATE SKIP LOCKED"
)

_DELETE_PROCESSED_QUERY = SQL("DELETE FROM {} " "WHERE id = %s")

_UPDATE_NON_PROCESSED_QUERY = SQL("UPDATE {} " "SET retry = retry + 1 " "WHERE id = %s")



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

from minos.common import (
    MinosConfig,
)

from ..abc import (
    Consumer,
)


class CommandConsumer(Consumer):
    """Command Consumer class."""

    TABLE_NAME = "command_queue"

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandConsumer:
        topics = [item.name for item in config.commands.items]
        return cls(topics=topics, broker=config.commands.broker, **config.commands.queue._asdict(), **kwargs)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from abc import (
    ABC,
)
from typing import (
    NoReturn,
)

from psycopg2.sql import (
    SQL,
)

from minos.common import (
    MinosBroker,
    PostgreSqlMinosDatabase,
)


class BrokerSetup(PostgreSqlMinosDatabase):
    """Minos Broker Setup Class"""

    async def _setup(self) -> NoReturn:
        await self._create_broker_table()

    async def _create_broker_table(self) -> NoReturn:
        await self.submit_query(_CREATE_TABLE_QUERY, lock=hash("producer_queue"))


class Broker(MinosBroker, BrokerSetup, ABC):
    """Minos Broker Class."""

    ACTION: str

    async def send_bytes(self, topic: str, raw: bytes) -> int:
        """Send a sequence of bytes to the given topic.

        :param topic: Topic in which the bytes will be send.
        :param raw: Bytes sequence to be send.
        :return: The identifier of the message in the queue.
        """
        params = (topic, raw, 0, self.ACTION)
        raw = await self.submit_query_and_fetchone(_INSERT_ENTRY_QUERY, params)
        return raw[0]


_CREATE_TABLE_QUERY = SQL(
    "CREATE TABLE IF NOT EXISTS producer_queue ("
    "id BIGSERIAL NOT NULL PRIMARY KEY, "
    "topic VARCHAR(255) NOT NULL, "
    "model BYTEA NOT NULL, "
    "retry INTEGER NOT NULL, "
    "action VARCHAR(255) NOT NULL, "
    "creation_date TIMESTAMP NOT NULL, "
    "update_date TIMESTAMP NOT NULL)"
)

_INSERT_ENTRY_QUERY = SQL(
    "INSERT INTO producer_queue (topic, model, retry, action, creation_date, update_date) "
    "VALUES (%s, %s, %s, %s, NOW(), NOW()) "
    "RETURNING id"
)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import logging
from typing import (
    NoReturn,
)

from minos.common import (
    Event,
    MinosConfig,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)

logger = logging.getLogger(__name__)


class EventHandler(Handler):
    """Event Handler class."""

    TABLE_NAME = "event_queue"
    ENTRY_MODEL_CLS = Event

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> EventHandler:
        handlers = {item.name: {"controller": item.controller, "action": item.action} for item in config.events.items}
        return cls(handlers=handlers, **config.events.queue._asdict(), **kwargs)

    async def dispatch_one(self, entry: HandlerEntry) -> NoReturn:
        """Dispatch one row.

        :param entry: Entry to be dispatched.
        :return: This method does not return anything.
        """
        logger.info(f"Dispatching '{entry.data!s}'...")
        await entry.callback(entry.topic, entry.data)



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import logging
from functools import (
    cached_property,
)
from inspect import (
    isawaitable,
    isclass,
)
from typing import (
    Awaitable,
    Callable,
    NoReturn,
    Optional,
    Union,
)

from aiohttp import (
    web,
)

from minos.common import (
    ENDPOINT,
    MinosConfig,
    MinosException,
    MinosSetup,
    Response,
    ResponseException,
    classname,
    import_module,
)

from .messages import (
    HttpRequest,
    HttpResponse,
)

logger = logging.getLogger(__name__)


class RestBuilder(MinosSetup):
    """
    Rest Interface Handler

    Rest Interface for aiohttp web handling.

    """

    def __init__(self, host: str, port: int, endpoints: list[ENDPOINT], **kwargs):
        super().__init__(**kwargs)
        self._host = host
        self._port = port
        self._endpoints = endpoints

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> RestBuilder:
        host = config.rest.broker.host
        port = config.rest.broker.port
        return cls(host=host, port=port, endpoints=config.rest.endpoints, **kwargs)

    @property
    def host(self) -> str:
        """Get the rest host.

        :return: A ``str`` object.
        """
        return self._host

    @property
    def port(self) -> int:
        """Get the rest port.

        :return: An integer value.
        """
        return self._port

    def get_app(self) -> web.Application:
        """Return rest application instance.

        :return: A `web.Application` instance.
        """
        return self._app

    @cached_property
    def _app(self) -> web.Application:
        app = web.Application()
        self._mount_routes(app)
        return app

    def _mount_routes(self, app: web.Application):
        """Load routes from config file."""
        for item in self._endpoints:
            self._mount_one_route(item, app)

        # Load default routes
        self._mount_system_health(app)

    def _mount_one_route(self, item: ENDPOINT, app: web.Application) -> NoReturn:
        action = self.get_action(item.controller, item.action)
        handler = self.get_callback(action)
        app.router.add_route(item.method, item.route, handler)

    @staticmethod
    def get_action(
        controller: str, action: str
    ) -> Callable[[HttpRequest], Union[Optional[HttpResponse], Awaitable[Optional[HttpResponse]]]]:
        """Load controller class and action method.
        :param controller: Controller string. Example: "tests.service.CommandTestService.CommandService"
        :param action: Config instance. Example: "get_order"
        :return: A class method callable instance.
        """
        controller = import_module(controller)
        if isclass(controller):
            controller = controller()
        action_fn = getattr(controller, action)
        return action_fn

    @staticmethod
    def get_callback(
        fn: Callable[[HttpRequest], Union[Optional[HttpResponse], Awaitable[Optional[HttpResponse]]]]
    ) -> Callable[[web.Request], Awaitable[web.Response]]:
        """Get the handler function to be used by the ``aiohttp`` Controller.

        :param fn: The action function.
        :return: A wrapper function around the given one that is compatible with the ``aiohttp`` Controller.
        """

        async def _fn(request: web.Request) -> web.Response:
            logger.info(f"Dispatching {classname(fn)!r} from {request.remote!r}...")
            request = HttpRequest(request)

            try:
                response = fn(request)
                if isawaitable(response):
                    response = await response
                if isinstance(response, Response):
                    response = await response.raw_content()
                if response is None:
                    return web.json_response()
                return web.json_response(response)
            except ResponseException as exc:
                logger.info(f"Raised a user exception: {exc!s}")
                raise web.HTTPBadRequest(text=str(exc))
            except MinosException as exc:
                logger.warning(f"Raised a 'minos' exception: {exc!r}")
                raise web.HTTPInternalServerError()
            except Exception as exc:
                logger.exception(f"Raised an exception: {exc!r}.")
                raise web.HTTPInternalServerError()

        return _fn

    def _mount_system_health(self, app: web.Application):
        """Mount System Health Route."""
        app.router.add_get("/system/health", self._system_health_handler)

    @staticmethod
    async def _system_health_handler(request: web.Request) -> web.Response:
        """System Health Route Handler.
        :return: A `web.json_response` response.
        """
        logger.info(f"Dispatching 'health' from {request.remote!r}...")
        return web.json_response({"host": request.host}, status=200)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from aiomisc.service.periodic import (
    PeriodicService,
)
from cached_property import (
    cached_property,
)

from minos.common import (
    PostgreSqlSnapshotBuilder,
)


class SnapshotService(PeriodicService):
    """Minos Snapshot Service class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_kwargs = kwargs

    async def start(self) -> None:
        """Start the service execution.

        :return: This method does not return anything.
        """
        await self.dispatcher.setup()
        await super().start()

    async def callback(self) -> None:
        """Callback implementation to be executed periodically.

        :return: This method does not return anything.
        """
        await self.dispatcher.dispatch()

    async def stop(self, err: Exception = None) -> None:
        """Stop the service execution.

        :param err: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        await super().stop(err)
        await self.dispatcher.destroy()

    @cached_property
    def dispatcher(self) -> PostgreSqlSnapshotBuilder:
        """Get the service dispatcher.

        :return: A ``PostgreSqlSnapshotBuilder`` instance.
        """
        return PostgreSqlSnapshotBuilder.from_config(**self._init_kwargs)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from aiohttp import (
    web,
)
from aiomisc import (
    bind_socket,
)
from aiomisc.service.aiohttp import (
    AIOHTTPService,
)
from cached_property import (
    cached_property,
)

from .builders import (
    RestBuilder,
)


class RestService(AIOHTTPService):
    """
    Rest Interface

    Expose REST Interface handler using aiomisc AIOHTTPService.

    """

    def __init__(self, **kwargs):
        super().__init__(**({"port": 9999} | kwargs))
        self._init_kwargs = kwargs

    async def create_application(self) -> web.Application:
        """Create the web application.

        :return: A ``web.Application`` instance.
        """
        return self.builder.get_app()  # pragma: no cover

    @cached_property
    def builder(self) -> RestBuilder:
        """Get the service builder.

        :return: A ``RestBuilder`` instance.
        """
        builder = RestBuilder.from_config(**self._init_kwargs)

        # Setup socket.
        self.socket = bind_socket(address=builder.host, port=builder.port, proto_name="http")

        return builder



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    Any,
)

from minos.common import (
    Command,
    Request,
    Response,
    ResponseException,
)


class CommandRequest(Request):
    """Command Request class."""

    __slots__ = "command"

    def __init__(self, command: Command):
        self.command = command

    def __eq__(self, other: CommandRequest) -> bool:
        return type(self) == type(other) and self.command == other.command

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.command!r})"

    async def content(self, **kwargs) -> Any:
        """Request content.

        :param kwargs: Additional named arguments.
        :return: The command content.
        """
        data = self.command.data
        return data


class CommandResponse(Response):
    """Command Response class."""


class CommandResponseException(ResponseException):
    """Command Response Exception class."""



# Copyright (C) 2020 Clariteia SL
#
# This file is part of minos framework.
#
# Minos framework can not be copied and/or distributed without the express
# permission of Clariteia SL.
from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    NoReturn,
)

from dependency_injector.wiring import (
    Provide,
)

from minos.common import (
    CommandReply,
    MinosConfig,
    MinosSagaManager,
)

from ..abc import (
    Handler,
)
from ..entries import (
    HandlerEntry,
)

logger = logging.getLogger(__name__)


class CommandReplyHandler(Handler):
    """Command Reply Handler class."""

    TABLE_NAME = "command_reply_queue"
    ENTRY_MODEL_CLS = CommandReply

    saga_manager: MinosSagaManager = Provide["saga_manager"]

    def __init__(self, saga_manager: MinosSagaManager = None, **kwargs: Any):
        super().__init__(**kwargs)

        if saga_manager is not None:
            self.saga_manager = saga_manager

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> CommandReplyHandler:
        handlers = {
            f"{item.name}Reply": {"controller": item.controller, "action": item.action} for item in config.saga.items
        }
        return cls(*args, handlers=handlers, **config.saga.queue._asdict(), **kwargs)

    async def dispatch_one(self, entry: HandlerEntry) -> NoReturn:
        """Dispatch one row.

        :param entry: Entry to be dispatched.
        :return: This method does not return anything.
        """
        logger.info(f"Dispatching '{entry.data!s}'...")
        await self.saga_manager.run(reply=entry.data)
